open Program 

exception Error of string

type loc = id
type value = Value of int | Addr of loc

module LocMap = Map.Make(struct type t = loc let compare = compare end)
module LblMap = Map.Make(struct type t = int let compare = compare end)

type memory = value LocMap.t
module MemSet = Set.Make(
  struct 
    type t = memory
    let compare = LocMap.compare (compare : value -> value -> int)
  end)

type collecting_semantics = MemSet.t LblMap.t

let string_of_value v =
  match v with
  | Value n -> string_of_int n
  | Addr a -> a

let string_of_mem mem =
  let first = ref true in
  let init = "[" in
  let f k v acc = 
    let semicolon = 
      if !first 
      then (first := false; "")
      else " ; " 
    in
    acc ^ semicolon ^ k ^ " |-> " ^ string_of_value v 
  in
  let ret = LocMap.fold f mem init in
  ret ^ "]"

let string_of_memset ms =
  let first = ref true in
  let init = "{" in
  let f v acc =
    let comma =
      if !first
      then (first := false; "\n ")
      else ",\n " 
    in
    acc ^ comma ^ string_of_mem v
  in
  let ret = MemSet.fold f ms init in
  if !first then ret ^ "}" else ret ^ "\n}"

let string_of_col col =
  let init = "" in
  let f k v acc = acc ^ string_of_int k ^ ":\n" ^ string_of_memset v ^ "\n" in
  LblMap.fold f col init

let example_mem : memory =
  LocMap.add "x" (Value 1) LocMap.empty

let example_mem1 : memory =
  LocMap.add "y" (Value 2) LocMap.empty

let example_memset : MemSet.t =
  MemSet.add example_mem1 (MemSet.singleton example_mem)

let example : collecting_semantics =
  LblMap.add 1 example_memset (LblMap.add 0 example_memset LblMap.empty)

(* let () = print_endline (string_of_col example) *)

(* d execution *)
let rec compute_exp : exp -> memory -> value = fun e mem -> (
  match e with 
  | NUM n -> Value n
  | VAR x -> LocMap.find x mem
  | DEREF x -> 
    let valu = LocMap.find x mem in (
    match valu with
    | Addr y -> LocMap.find y mem
    | _ -> raise (Error ("address expected for dereference, but got integer")))
  | AT x -> Addr x
  | ADD (e1, e2) -> 
    let x1 = compute_exp e1 mem in
    let x2 = compute_exp e2 mem in (
      match x1 with
      | Value n1 -> (
        match x2 with
        | Value n2 -> Value (n1 + n2)
        | _ -> raise (Error ("integer expected for second operand of add, but got address"))
      )
      | _ -> raise (Error "integer expected for first operand of add, but got address"))
  | MULT (e1, e2) ->
    let x1 = compute_exp e1 mem in
    let x2 = compute_exp e2 mem in (
      match x1 with
      | Value n1 -> (
        match x2 with
        | Value n2 -> Value (n1 * n2)
        | _ -> raise (Error ("integer expected for second operand of mult, but got address"))
      )
      | _ -> raise (Error "integer expected for first operand of mult, but got address")
    )
  | MINUS e1 ->
    let x1 = compute_exp e1 mem in (
      match x1 with
      | Value n1 -> Value (-n1)
      | _ -> raise (Error ("integer expected for operand of minus, but got address"))
    )
  | TRUE -> Value 1
  | FALSE -> Value 0
  | READ -> Value (read_int ())
  | NOT e1 ->
    let x1 = compute_exp e1 mem in (
      match x1 with
      | Value n1 -> Value (if n1 = 0 then 1 else 0)
      | _ -> raise (Error ("integer expected for operand of not, but got address"))
    )
  | ANDALSO (e1, e2) ->
    let x1 = compute_exp e1 mem in
    let x2 = compute_exp e2 mem in (
      match x1 with
      | Value n1 -> (
        match x2 with
        | Value n2 -> Value (if (n1 = 0 || n2 = 0) then 0 else 1)
        | _ -> raise (Error ("integer expected for second operand of and, but got address"))
      )
      | _ -> raise (Error ("integer expected for first operand of and, but got address"))
    )
  | ORELSE (e1, e2) ->
    let x1 = compute_exp e1 mem in
    let x2 = compute_exp e2 mem in (
      match x1 with
      | Value n1 -> (
        match x2 with
        | Value n2 -> Value (if (n1 != 0 || n2 != 0) then 1 else 0)
        | _ -> raise (Error ("intger expected for second operand of or, but got address"))
      )
      | _ -> raise (Error ("integer expected for first operand of or, but got address"))
    )
  | LESS (e1, e2) -> (
    let x1 = compute_exp e1 mem in
    let x2 = compute_exp e2 mem in (
      match x1 with
      | Value n1 -> (
        match x2 with
        | Value n2 -> Value (if n1 < n2 then 1 else 0)
        | _ -> raise (Error ("integer expected for second operand of less, but got address"))
      )
      | _ -> raise (Error ("integer expected for first operand of less, but got address"))
    )
  )
)

(* 
Given a program with label, compute next(l), nextTrue(l), nextFalse(l) for each label l.
f(0, l) means next(l)
f(1, l) means nextTrue(l)
f(2, l) means nextFalse(l)
*)

type next_f = (int * int) -> int
let rec compute_next : program -> int -> next_f = fun pgm l' -> (
  match pgm with
  (* Label of the current program is l *)
  | LBL (l, c) -> (
		match c with
		| SEQ (p1, p2) -> (
			match p1 with
			| LBL (l1, c1) -> (
				match p2 with 
				| LBL (l2, c2) -> (
					let mp1 = compute_next p1 l2 in
					let mp2 = compute_next p2 l' in 
					(fun x -> if x = (0, l) then l1 else (if (mp1 x) != -1 then (mp1 x) else (mp2 x)))
				)
			)
		)
		| IF (_, p1, p2) -> (
			match p1 with
			| LBL (l1, c1) -> (
				match p2 with
				| LBL (l2, c2) -> (
					let mp1 = compute_next p1 l' in
					let mp2 = compute_next p2 l' in
					(fun x -> 
						if x = (1, l) then l1 else 
						if x = (2, l) then l2 else 
						if (mp1 x) != -1 then (mp1 x) else (mp2 x))
				)
			)
		)
		| WHILE (_, p1) -> (
			match p1 with
			| LBL (l1, c1) -> (
				let mp1 = compute_next p1 l in
				(fun x ->
					if x = (1, l) then l1 else
					if x = (2, l) then l' else (mp1 x))
			)
		)
		| ASSIGN (_, _) | PTRASSIGN (_, _) -> (fun x -> if x = (0, l) then l' else -1)
		| _ -> (fun x -> -1)
  )
)

(*
Given a program and its label, finds and returns the sub-program with given label	 
*)
let rec find_program : program -> int -> program = fun pgm l -> (
  match pgm with
  | LBL (l', c) -> (
    if l = l' then pgm else (
      match c with
      | SEQ (p1, p2) | IF (_, p1, p2) -> 
        let r1 = find_program p1 l in
        let r2 = find_program p2 l in (
        match r1 with 
        | LBL (l1, _) -> if l1 = l then r1 else r2
      )
      | WHILE (_, p1) -> find_program p1 l
      | _ -> pgm
    )
	)
)

(*
Given current (label, memory), next() function and the whole program, proceed by one step	 
*)
type state = int * memory

let rec step : state -> next_f -> program -> state = fun (l, m) next pgm -> (
	let p = find_program pgm l in (
		match p with
		| LBL (l', c) -> (
			if l != l' then raise (Error ("invalid label")) else (
				match c with
				| SEQ (p1, p2) -> (next (0, l), m)
				| IF (e, _, _) | WHILE (e, _) ->
					let v = compute_exp e m in (
						match v with
						| Value n -> if (n == 0) then (next (2, l), m) else (next (1, l), m)
						| _ -> raise (Error ("expected integer for if/while-condition, but got address"))
					)
				| GOTO e -> 
					let v = compute_exp e m in (
						match v with
						| Value n -> (n, m)
						| _ -> raise (Error ("expected integer for goto, but got address"))
					)
				| ASSIGN (x, e) ->
					let v = compute_exp e m in (next (0, l), LocMap.add x v m)
				| PTRASSIGN (x, e) -> (
					let v = compute_exp e m in 
					let px = LocMap.find x m in (
						match px with
						| Addr y -> (next (0, l), LocMap.add y v m)
						| _ -> raise (Error ("expected address for ptrassign, but got integer"))
					)
				)
			)
		)
	)
)
let rec steps : state -> next_f -> program -> state = fun (l, m) next pgm -> (
	if l = -1 then (l, m) else steps (step (l, m) next pgm) next pgm
)
let rec dintp : program -> memory = fun pgm ->
	let next = compute_next pgm (-1) in
	let (l, m) = steps (0, LocMap.empty) next pgm in
	m

(*
Given current (label, memory), next() function, whole program and the collecting set,
proceed by one step and return new label and memory, new collecting set	 
*)

type c_state = (int * memory) * collecting_semantics

let rec collect_step : c_state -> next_f -> program -> c_state = fun ((l, m), st) next pgm -> (
	let (l', m') = step (l, m) next pgm in	
	let st' = LblMap.add l (MemSet.add m (LblMap.find l st)) st in
	((l', m'), st')
)
let rec collect_steps : c_state -> next_f -> program -> c_state = fun ((l, m), st) next pgm -> (
	if l = -1 then ((l, m), st) else collect_steps (collect_step ((l, m), st) next pgm) next pgm
)

let rec init : program -> collecting_semantics -> collecting_semantics = fun pgm st -> (
	match pgm with
	| LBL (l, c) -> (
		let st' = LblMap.add l (MemSet.empty) st in (
			match c with
			| SEQ (p1, p2) | IF (_, p1, p2) ->
				let st1 = init p1 st' in init p2 st1
			| WHILE (_, p1) -> init p1 st'
			| _ -> st'
		)
	)
)
(* d analysis *)
let rec dcintp : program -> collecting_semantics = fun pgm ->
	let next = compute_next pgm (-1) in
	let st = init pgm (LblMap.empty) in
	let ((l, m), st) = collect_steps ((0, LocMap.empty), st) next pgm in
	st
